from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Iterator, Literal

from ..comic import BaseComic

ACCEPTED_IMAGE_EXTENSIONS = {
    ".gif": "gif",
    ".png": "png",
    ".jpg": "jpeg",
    ".jpeg": "jpeg",
    ".jpe": "jpeg",
}


class ConvertFormats(str, Enum):
    # for typing purposes
    CBZ = "cbz"
    EPUB = "epub"
    PDF = "pdf"
    NONE = "none"


@dataclass(kw_only=True, slots=True)
class ConvertOptions:
    page_progression: Literal["rtl"] | Literal["ltr"] = "ltr"


class BaseConverter:
    def __init__(self, comic: BaseComic, options: ConvertOptions | None = None) -> None:
        self.comic = comic
        self.options = options or ConvertOptions()

        self.__post_init__()

    def __post_init__(self) -> None:
        pass

    def create_file_progress(
        self, path: Path | str, save_to: Path | str
    ) -> Iterator[str]:
        raise NotImplementedError

    def create_file(self, path: Path | str, save_to: Path | str) -> None:
        for _ in self.create_file_progress(path, save_to):
            pass



# pylint: disable=no-member
"""
I'M SORRY I DON'T UNDERSTAND EPUB OKAY
"""

import shutil
import tempfile
import zipfile
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Iterator
from uuid import uuid4

from lxml import etree
from lxml.builder import E, ElementMaker
from lxml.etree import QName

from ..io import NUM_LEFT_PAD_DIGITS, discover_local_images
from .base_converter import ACCEPTED_IMAGE_EXTENSIONS, BaseConverter

STYLE_CSS = """
@page {
    margin: 0
}

body {
    display: block;
    margin: 0;
    padding: 0;
}
""".strip()

CONTAINER_XML = """
<?xml version="1.0"?>
<container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">
<rootfiles>
<rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/>
</rootfiles>
</container>
""".strip()

OPF_MAP = {
    "opf": "http://www.idpf.org/2007/opf",
    "dc": "http://purl.org/dc/elements/1.1/",
}

NAV_MAP = {
    "epub": "http://www.idpf.org/2007/ops",
    None: "http://www.w3.org/1999/xhtml",
}


class EpubConverter(BaseConverter):
    def __post_init__(self) -> None:
        self.uuid = str(uuid4())

    def create_file_progress(
        self, path: Path | str, save_to: Path | str
    ) -> Iterator[str]:
        path = Path(path)
        save_to = Path(save_to)
        slug_map = discover_local_images(path)

        with tempfile.TemporaryDirectory() as temp:
            root = Path(temp)
            oebps = self.create_skeleton(root)
            (oebps / "toc.ncx").write_text(self.toc_ncx())
            (oebps / "nav.xhtml").write_text(self.nav_xhtml())
            (oebps / "content.opf").write_text(self.content_opf(slug_map))
            (oebps / "Text" / "style.css").write_text(STYLE_CSS)

            for chap in self.comic.chapters:
                yield "Building"

                text_dir = oebps / "Text" / chap.slug
                text_dir.mkdir(exist_ok=True)

                image_dir = oebps / "Images" / chap.slug
                image_dir.mkdir(exist_ok=True)

                for _, image in enumerate((path / chap.slug).iterdir(), start=1):
                    if image.suffix in ACCEPTED_IMAGE_EXTENSIONS:
                        new_file = text_dir / f"{image.with_suffix('').name}.xhtml"
                        new_file.write_text(self.generate_image_html(chap.slug, image))

                        shutil.copyfile(
                            image,
                            image_dir / image.name,
                        )

            dest_file = save_to / f"{self.comic.metadata.title}.epub"
            with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as zipout:
                zipout.writestr("mimetype", "application/epub+zip", zipfile.ZIP_STORED)
                for p in root.rglob("*"):
                    yield "Compressing"
                    if p.is_file():
                        relative_path = Path(str(p).lstrip(str(root)))
                        zipout.write(p, relative_path, zipfile.ZIP_DEFLATED)

    def create_skeleton(self, tmpdir: Path) -> Path:
        """
        Create the following structure at `tmpdir`:
        ```
        tmpdir/
          - META-INF/container.xml
          - OEBPS/
              - Images/
              - Text/
        ```
        :return The OEBPS directory.
        """
        meta_inf = tmpdir / "META-INF"
        meta_inf.mkdir()
        container_xml = meta_inf / "container.xml"
        container_xml.write_text(CONTAINER_XML, encoding="utf-8")

        oebps = tmpdir / "OEBPS"
        image_dir = oebps / "Images"
        text_dir = oebps / "Text"

        text_dir.mkdir(parents=True)
        image_dir.mkdir(parents=True)

        return oebps

    def generate_image_html(
        self,
        chapter_slug: str,
        image_path: Path,
    ) -> str:
        X = ElementMaker(  # pylint: disable=invalid-name
            namespace="http://www.w3.org/1999/xhtml",
            nsmap=NAV_MAP,
        )
        image_base = str(image_path.with_suffix(""))
        tree = X.html(
            E.head(
                E.title(image_base),
                E.link(href="../style.css", type="text/css", rel="stylesheet"),
            ),
            E.body(
                E.div(style="text-align:center;top:0.0%"),
                E.img(
                    src=f"../../Images/{chapter_slug}/{image_path.name}",
                ),
            ),
        )

        return etree.tostring(
            tree,
            xml_declaration=True,
            encoding="utf-8",
            pretty_print=True,
        ).decode("utf-8")

    def toc_ncx(self) -> str:
        nav_map: list[etree._Element] = []

        for chap in self.comic.chapters:
            nav_map.append(
                E.navPoint(
                    E.navLabel(
                        E.text(chap.title),
                    ),
                    E.content(src=f"Text/{chap.slug}/{1:0{NUM_LEFT_PAD_DIGITS}}.xhtml"),
                    id=f"Text_{chap.slug}",
                )
            )

        tree = E.ncx(
            E.head(
                E.meta(name="dtb:uid", content=f"urn:uuid:{self.uuid}"),
                E.meta(name="dtb:depth", content="1"),
                E.meta(name="dtb:totalPageCount", content="0"),
                E.meta(name="dtb.maxPageNumber", content="0"),
                E.meta(name="generated", content="true"),
            ),
            E.docTitle(E.text(self.comic.metadata.title)),
            E.navMap(*nav_map),
            version="2005-1",
            xmlns="http://www.daisy.org/z3986/2005/ncx/",
        )
        tree.attrib[QName("http://www.w3.org/XML/1998/namespace", "lang")] = "en-US"

        return etree.tostring(
            tree,
            xml_declaration=True,
            encoding="utf-8",
            pretty_print=True,
        ).decode("utf-8")

    def content_opf(
        self, slug_map: dict[str, list[Path]], cover: Path | None = None
    ) -> str:
        metadata = self.comic.metadata

        item_ids: list[etree._Element] = []
        item_img_ids: list[etree._Element] = [
            E.item(
                id="nav",
                href="nav.xhtml",
                properties="nav",
                **{"media-type": "application/xhtml+xml"},
            ),
            E.item(
                id="ncx", href="toc.ncx", **{"media-type": "application/x-dtbncx+xml"}
            ),
            E.item(id="css", href="Text/style.css", **{"media-type": "text/css"}),
        ]
        item_refs: list[etree._Element] = []
        for _, chap in enumerate(self.comic.chapters, start=1):
            for index, image in enumerate(slug_map[chap.slug], start=1):
                ref_id = (
                    f"page_Images_C{chap.slug}-{index:0{NUM_LEFT_PAD_DIGITS}}-mandown"
                )
                image_ref_id = (
                    f"img_Images_C{chap.slug}-{index:0{NUM_LEFT_PAD_DIGITS}}-mandown"
                )

                img_el = E.item(
                    id=image_ref_id,
                    href=f"Images/{chap.slug}/{image.name}",
                    **{
                        "media-type": f"image/{ACCEPTED_IMAGE_EXTENSIONS[image.suffix]}"
                    },
                )

                text_el = E.item(
                    id=ref_id,
                    href=f"Text/{chap.slug}/{image.with_suffix('').name}.xhtml",
                    **{"media-type": "application/xhtml+xml"},
                )

                item_ids.append(text_el)
                item_img_ids.append(img_el)
                item_refs.append(E.itemref(idref=ref_id))

        if cover:
            cover_el = E.item(
                id="cover",
                href=f"Images/{cover.name}",
                properties="cover-image",
                **{"media-type": f"image/{ACCEPTED_IMAGE_EXTENSIONS[cover.suffix]}"},
            )
            item_ids.append(cover_el)

        M = ElementMaker(  # pylint: disable=invalid-name
            # namespace="http://purl.org/dc/elements/1.1/",
            nsmap=OPF_MAP,
        )

        MD = ElementMaker(  # pylint: disable=invalid-name
            namespace="http://purl.org/dc/elements/1.1/",
            nsmap=OPF_MAP,
        )

        els_genres = [MD.subject(i) for i in metadata.genres]
        els_authors = [MD.creator(i) for i in metadata.authors]

        time_now = datetime.now().replace(microsecond=0).isoformat()

        package = E.package(
            M.metadata(
                MD.title(metadata.title),
                *els_authors,
                MD.contributor("mandown"),
                MD.description(metadata.description),
                MD.identifier(metadata.url),
                MD.identifier(f"urn:uuid:{self.uuid}", id="BookID"),
                MD.language("eng"),
                *els_genres,
                E.meta(name="mandown:cover", content=metadata.cover_art),
                E.meta(f"{time_now}Z", property="dcterms:modified"),
                # E.guide(E.reference(type="cover", title="Cover", href="cover")),
            ),
            E.manifest(*item_ids, *item_img_ids),
            E.spine(
                *item_refs,
                toc="ncx",
                **{"page-progression-direction": "ltr"},
            ),
            xmlns=OPF_MAP["opf"],
            version="3.0",
            **{"unique-identifier": "BookID"},
        )

        return etree.tostring(
            package,
            xml_declaration=False,
            encoding="utf-8",
            pretty_print=True,
        ).decode("utf-8")

    def nav_xhtml(self) -> str:
        X = ElementMaker(  # pylint: disable=invalid-name
            namespace="http://www.w3.org/1999/xhtml",
            nsmap=NAV_MAP,
        )

        els_list: list[etree._Element] = []
        for _, chap in enumerate(self.comic.chapters, start=1):
            href = f"Text/{chap.slug}/{1:0{NUM_LEFT_PAD_DIGITS}}.xhtml"
            els_list.append(E.li(E.a(chap.title, href=href)))

        # you can't reuse elements after much head-banging
        els_list2 = deepcopy(els_list)

        tree = X.html(
            E.head(
                E.title(self.comic.metadata.title),
                E.meta(charset="utf-8"),
            ),
            E.body(
                X.nav(
                    E.ol(*els_list2),
                    id="toc",
                    **{"{http://www.idpf.org/2007/ops}type": "toc"},
                ),
                X.nav(
                    E.ol(*els_list),
                    **{"{http://www.idpf.org/2007/ops}type": "page-list"},
                ),
            ),
        )

        return etree.tostring(
            tree,
            xml_declaration=True,
            encoding="utf-8",
            pretty_print=True,
        ).decode("utf-8")


def get_class() -> type[EpubConverter]:
    return EpubConverter



import json
import multiprocessing as mp
import os
import urllib.parse
from pathlib import Path
from typing import Iterator, Sequence

import filetype
import requests
from natsort import natsorted

from . import sources
from .base import BaseChapter, BaseMetadata
from .comic import BaseComic

NUM_LEFT_PAD_DIGITS = 5
FILE_PADDING = f"0{NUM_LEFT_PAD_DIGITS}"
MD_METADATA_FILE = "md-metadata.json"


def async_download_image(
    data: tuple[str, Path | str, str | None, dict[str, str] | None]
) -> None:
    url, dest_folder, filename, headers = data
    dest_folder = Path(dest_folder)

    name = filename or url.split("/")[-1]
    dest_file = dest_folder / name

    res = requests.get(url, headers=headers)
    res.raise_for_status()
    with open(dest_file, "wb") as file:
        file.write(res.content)

    # if the file extension is lying
    # rename it so epubcheck doesn't yell at us
    ext = filetype.guess(dest_file)
    if ext is not None and ext.extension in ["jpg", "png", "gif"]:
        dest_file.rename(dest_file.with_suffix(f".{ext.extension}"))


def download_images(
    urls: Sequence[str],
    dest_folder: Path | str,
    *,
    filestems: Sequence[str] | None = None,
    headers: dict[str, str] | None = None,
    threads: int = 1,
) -> Iterator[None]:
    """
    Download one or multiple URLs to a destination folder.
    Raises ValueError if the folder does not exist.
    :param `urls`: A list of URLs to download.
    :param `dest_folder`: The path to download files into.
    :param `filestems`: Specify the name of each downloaded file instead of the default.
    :param `headers`: Request headers
    :param `threads`: The number of processes to open
    """
    dest_folder = Path(dest_folder)

    # attempt to create
    dest_folder.mkdir(exist_ok=True)

    # args to async_download
    map_pool: list[tuple[str, Path | str, str | None, dict[str, str] | None]] = []

    if filestems is None:
        filestems = [f"{i+1:FILE_PADDING}" for i in range(len(urls))]

    for url, stem in zip(urls, filestems, strict=True):
        _, ext = os.path.splitext(urllib.parse.urlparse(url).path)
        map_pool.append((url, dest_folder, f"{stem}{ext}", headers))

    with mp.Pool(threads) as pool:
        yield from pool.imap_unordered(async_download_image, map_pool)


def read_comic(path: Path | str) -> BaseComic:
    """
    Open a comic from a folder path.
    :param `path`: A folder containing `md-metadata.json`
    :returns A comic with metadata and chapter data of that folder
    :raises `FileNotFoundError` if `md-metadata.json` is not found
    """
    path = Path(path)
    json_path = path / MD_METADATA_FILE

    with open(json_path, "r", encoding="utf-8") as file:
        data = json.load(file)

    return BaseComic(
        BaseMetadata(**data["metadata"]),
        [BaseChapter(**c) for c in data["chapters"]],
    )


def parse_comic(path: Path | str, source_url: str | None = None) -> BaseComic:
    """
    Parse and return an incomplete comic (without most metadata)
    :param `url`: A source URL to fill metadata from
    """
    path = Path(path)

    title = path.stem
    authors: list[str] = []
    url = ""
    genres: list[str] = []
    description = ""
    cover_art = ""
    metadata = BaseMetadata(title, authors, url, genres, description, cover_art)

    chapters = [
        BaseChapter(inode.stem, "", inode.stem)
        for inode in natsorted(path.iterdir(), key=lambda i: i.stem)
        if inode.is_dir()
    ]

    if source_url:
        new_comic = sources.get_class_for(source_url)(source_url)
        metadata = new_comic.metadata
        for local, remote in zip(chapters, new_comic.chapters):
            local.title = remote.title
            local.url = remote.url

    return BaseComic(metadata, chapters)


def save_comic(comic: BaseComic, path: Path | str) -> None:
    """
    Save an `md-metadata.json` from `comic` into `path`.
    """
    path = Path(path)
    path.mkdir(exist_ok=True)

    json_path = path / MD_METADATA_FILE

    with open(json_path, "w", encoding="utf-8") as file:
        json.dump(comic.asdict(), file)


def discover_local_images(path: Path | str) -> dict[str, list[Path]]:
    """
    Given a comic path, return a dictionary of slugs: images.
    Basically a slightly modified version of os.walk.
    """
    path = Path(path)

    return {
        chap.stem: list(sorted(chap.iterdir()))
        for chap in sorted(path.iterdir())  # iterdir does not guarantee any order
        if chap.is_dir()  # force explosion for readability
    }



#!/usr/bin/env python3

from pathlib import Path
from typing import List, Optional

import typer

from . import __version_str__, api, sources
from .comic import BaseComic
from .converter.base_converter import ConvertFormats
from .io import MD_METADATA_FILE
from .processor import ProcessOps

app = typer.Typer()


def cli_query(url: str) -> BaseComic:
    typer.echo(f"Searching sources for {url}")

    try:
        comic = api.query(url)
    except ValueError as err:
        typer.secho("Could not match URL with available sources.", fg=typer.colors.RED)
        raise typer.Exit(1) from err
    return comic


def cli_convert(
    comic_path: Path, target_format: ConvertFormats, dest_folder: Path = Path.cwd()
) -> None:
    try:
        comic = api.load(comic_path)
    except FileNotFoundError as err:
        typer.secho(
            f"Comic not found at {comic_path}, is md-metadata.json missing?",
            fg=typer.colors.RED,
        )
        raise typer.Exit(1) from err

    with typer.progressbar(
        api.convert_progress(comic, comic_path, target_format, dest_folder),
        length=len(comic.chapters),
        label="Converting",
    ) as progress:
        for _ in progress:
            pass
    dest_file = dest_folder / f"{comic.metadata.title}.{target_format.value}"
    typer.secho(f"Successfully converted to {dest_file}", fg=typer.colors.GREEN)


def cli_process(comic_path: Path, options: list[ProcessOps]) -> None:
    if ProcessOps.NO_POSTPROCESSING in options:
        return

    try:
        comic = api.load(comic_path)
    except FileNotFoundError as err:
        typer.secho(
            f"Comic not found at {comic_path}, is md-metadata.json missing?",
            fg=typer.colors.RED,
        )
        raise typer.Exit(1) from err

    typer.secho(
        f"Applying processing options: {', '.join(options)}", fg=typer.colors.GREEN
    )
    with typer.progressbar(
        api.process_progress(comic_path, options),
        length=len(comic.chapters),
        label="Processing",
    ) as progress:
        for _ in progress:
            pass


@app.command()
def convert(
    convert_to: ConvertFormats,
    folder_path: Path,
    dest: Path = typer.Option(
        Path.cwd(),
        "--dest",
        "-d",
        help="The folder to save the converted file to.",
    ),
) -> None:
    """
    Convert a comic folder into CBZ/EPUB/PDF.
    """
    try:
        comic = api.load(folder_path)
    except FileNotFoundError as err:
        typer.secho(
            f"Comic not found at {folder_path}, is md-metadata.json missing?",
            fg=typer.colors.RED,
        )
        raise typer.Exit(1) from err

    typer.echo(
        f"Found {comic.metadata.title} with {len(comic.chapters)} chapters, "
        f"converting to {convert_to}..."
    )
    cli_convert(folder_path, convert_to, dest)


@app.command()
def process(
    options: list[ProcessOps], folder_path: Path = typer.Argument(Path.cwd())
) -> None:
    """
    Process a comic folder in-place.
    """
    cli_process(folder_path, options)


@app.command()
def get(
    url: str,
    dest: Path = typer.Argument(
        Path.cwd(), help="The destination folder to download to."
    ),
    convert_to: ConvertFormats = typer.Option(
        "none", "--convert", "-c", help="The format to download the comic as"
    ),
    start: Optional[int] = typer.Option(
        None,
        "--start",
        "-s",
        help="The first chapter to download [default: first found]",
    ),
    end: Optional[int] = typer.Option(
        None, "--end", "-e", help="The last chapter to download [default: last found]"
    ),
    maxthreads: int = typer.Option(
        4,
        "--threads",
        "-t",
        help="The maximum number of images to download in parallel",
    ),
    processing_options: Optional[List[ProcessOps]] = typer.Option(
        [],
        "--processing-options",
        "-p",
        help="Image processing options (in-place)",
        case_sensitive=True,
    ),
) -> None:
    """
    Download from a URL chapters start_chapter to end_chapter.
    Defaults to the first chapter and last chapter, respectively
    in the working directory.
    """
    if not dest.is_dir():
        raise ValueError(f"{dest} is not a valid folder path.")

    # get and save metadata
    comic = cli_query(url)

    typer.echo(
        f"Found item from source {sources.get_class_for(url).name}",
    )

    # get processing range
    start_chapter = start or 1
    end_chapter = end or len(comic.chapters)

    # zero-index
    start_chapter -= 1

    comic.set_chapter_range(start=start_chapter, end=end_chapter)

    # download
    typer.echo("Downloading...")
    with typer.progressbar(
        api.download_progress(comic, dest, threads=maxthreads),
        length=len(comic.chapters),
    ) as progress:
        for title in progress:
            progress.label = title
    typer.secho(
        f"Successfully downloaded {end_chapter - start_chapter} chapters.",
        fg=typer.colors.GREEN,
    )

    # process
    if processing_options:
        cli_process(dest / comic.metadata.title, processing_options)

    # convert
    if convert_to != ConvertFormats.NONE:
        cli_convert(dest / comic.metadata.title, convert_to, dest)


@app.command(name="init-metadata")
def init_metadata(
    path: Path,
    source_url: Optional[str] = typer.Argument(
        None, help="The url to get metadata from"
    ),
) -> None:
    """
    Initialise a folder with metadata to be converted with Mandown, optionally
    fetching metadata from an internet source.
    """
    if (path / MD_METADATA_FILE).is_file():
        return typer.echo(
            "Metadata already found. Please remove it to create new metadata."
        )
    comic = api.init_parse_comic(path, source_url)
    typer.secho(f"Found {comic.metadata.title}:", fg=typer.colors.BRIGHT_GREEN)
    typer.echo(comic)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def callback(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        is_eager=True,
        help="Display the current version of mandown",
    ),
    supported_sites: Optional[bool] = typer.Option(
        None,
        "--supported-sites",
        is_eager=True,
        help="Output a list of domains supported by mandown",
    ),
) -> None:
    if version:
        typer.echo(f"mandown {__version_str__}")
        raise typer.Exit()

    if supported_sites:
        for source in sources.get_all_classes():
            typer.echo(f"{source.name}: {', '.join(source.domains)}")
        raise typer.Exit()


def main() -> None:
    app()


if __name__ == "__main__":
    main()



import zipfile
from pathlib import Path
from typing import Iterator

from .base_converter import BaseConverter


class CbzConverter(BaseConverter):
    def create_file_progress(
        self, path: Path | str, save_to: Path | str
    ) -> Iterator[str]:
        path = Path(path).absolute()
        save_to = Path(save_to)

        dest_file = save_to / f"{self.comic.metadata.title}.cbz"

        with zipfile.ZipFile(dest_file, "w", zipfile.ZIP_DEFLATED) as file:
            for outerpath in path.iterdir():
                if outerpath.is_dir():
                    for image in outerpath.iterdir():
                        yield "Compressing"
                        file.write(
                            image,
                            image.absolute().relative_to(path),
                            zipfile.ZIP_DEFLATED,
                        )
                elif outerpath.stem == "cover":
                    file.write(
                        outerpath,
                        # rename cover to appear first in the archive
                        outerpath.with_stem("__cover").absolute().relative_to(path),
                        zipfile.ZIP_DEFLATED,
                    )


def get_class() -> type[CbzConverter]:
    return CbzConverter



from pathlib import Path
from typing import Iterator

from .base_converter import ACCEPTED_IMAGE_EXTENSIONS, BaseConverter

try:
    from PIL import Image

    HAS_PILLOW = True
except ImportError:
    HAS_PILLOW = False


PDF_IMAGE_MAX_INTERVAL = 200  # adjust for memory as necessary
PDF_IMAGE_MIN_INTERVAL_FACTOR = 0.12


class PdfConverter(BaseConverter):
    def create_file_progress(
        self, path: Path | str, save_to: Path | str
    ) -> Iterator[str]:
        if not HAS_PILLOW:
            raise ImportError(
                "Pillow could not be found and is required for PDF conversion"
            )

        path = Path(path)
        save_to = Path(save_to)

        images: list[Image.Image] = [
            Image.open(f)
            for f in sorted(Path.rglob(path, "*"))
            if f.suffix in ACCEPTED_IMAGE_EXTENSIONS
        ]

        if len(images) < 0:
            raise IOError("No images to convert found")

        # interval for num pdf images to process
        # at one time
        interval = round(
            min(
                PDF_IMAGE_MAX_INTERVAL,
                len(images) * PDF_IMAGE_MIN_INTERVAL_FACTOR,
            )
        )

        dest_file = save_to / (self.comic.metadata.title + ".pdf")
        for i in range(0, len(images), interval):
            append_images = (
                images[i + 1 : i + interval] if len(images) > i + 1 else None
            )
            author = (
                self.comic.metadata.authors[0] if self.comic.metadata.authors else ""
            )

            images[i].save(
                dest_file,
                "PDF",
                resolution=100.0,
                save_all=True,
                append_images=append_images,
                title=self.comic.metadata.title,
                author=author,
                append=i != 0,
                creator="mandown",
                producer="mandown",
            )

            for j in images[i : i + interval]:
                j.close()
            yield "Building"


def get_class() -> type[PdfConverter]:
    return PdfConverter



from typing import TYPE_CHECKING

try:
    from PIL import Image, ImageChops
except ImportError:

    if not TYPE_CHECKING:

        class Image:
            class Image:
                pass

        class ImageChops:
            pass


class ProcessContainer:
    """
    Add processing functions here!
    Name them by their Enum string in ProcessOps.
    They should all return a Image.Image | tuple[Image.Image] | None
    depending on if they can return multiple images.

    The first image returned will replace the current image,
    all other ones will be suffixed with `a`. If `None` is returned,
    the image will not be altered.
    """

    @staticmethod
    def rotate_double_pages(image: Image.Image) -> Image.Image | None:
        width, height = image.size
        if width > height:
            return image.rotate(90, expand=1)
        return None

    @staticmethod
    def split_double_pages(
        image: Image.Image,
    ) -> tuple[Image.Image, Image.Image] | None:
        width, height = image.size
        if not width > height:
            return None

        left = image.crop((0, 0, int(width / 2), height))
        right = image.crop((int(width / 2), 0, width, height))
        return (left, right)

    @staticmethod
    def trim_borders(image: Image.Image) -> Image.Image | None:
        bg = Image.new(image.mode, image.size, image.getpixel((0, 0)))
        diff = ImageChops.difference(image, bg)
        diff = ImageChops.add(diff, diff, 2.0, -100)
        bbox = diff.getbbox()
        if bbox:
            return image.crop(bbox)
        return None



from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

from .ops import ProcessContainer

try:
    from PIL import Image

    HAS_PILLOW = True
except ImportError:
    HAS_PILLOW = False

    if not TYPE_CHECKING:

        class Image:
            class Image:
                pass


class ProcessOps(str, Enum):
    ROTATE_DOUBLE_PAGES = "rotate_double_pages"
    """If page width is greater than its height, rotate it 90 degrees."""
    SPLIT_DOUBLE_PAGES = "split_double_pages"
    """If page width is greater than its height, split it in half into two images."""
    NO_POSTPROCESSING = "none"
    """Disable image processing entirely."""
    TRIM_BORDERS = "trim_borders"
    """Conservatively remove borders in images."""


class Processor:
    def __init__(self, image_path: Path | str) -> None:
        if not HAS_PILLOW:
            raise ImportError(
                "Pillow was not found and is needed for processing. Is it installed?"
            )

        self.image_path = Path(image_path)
        self._image = Image.open(self.image_path)
        self.is_modified = False

        # dangerous if there are multiple types of operations
        # that would add new image files to be written
        self.new_images: list[Image.Image] = []

    @property
    def image(self) -> Image.Image:
        return self._image

    @image.setter
    def image(self, image: Image.Image) -> None:
        self._image = image
        self.is_modified = True

    def write(self, filename: Path | str | None = None) -> None:
        """Save the processed image(s) manually"""
        filename = Path(filename or self.image_path)
        self.image.save(filename)

        for image in self.new_images:
            # increment by one "a" each time
            filename = filename.with_stem(filename.stem + "a")
            image.save(filename)

    def process(
        self, operations: list[ProcessOps], filename: Path | str | None = None
    ) -> None:
        """
        Perform the operations in `operations` on the image in sequence
        and save it to disk. If `filename` is not None, it will be saved
        with that filename instead.
        """
        if ProcessOps.NO_POSTPROCESSING in operations:
            return

        for func in operations:
            try:
                images: tuple[Image.Image, ...] | Image.Image | None = getattr(
                    ProcessContainer, func
                )(self.image)

                if images is None:
                    continue

                if isinstance(images, Image.Image):
                    images = (images,)

                self.image = images[0]

                if len(images) > 1:
                    self.new_images.extend(images[1:])

                self.is_modified = True
            except AttributeError as err:
                raise NotImplementedError(
                    f"{func} is not a valid post-processing function."
                ) from err
            except OSError as err:
                raise OSError(f"Error in {self.image_path}") from err

        if self.is_modified:
            # only write to disk if something has actually changed
            self.write(filename)



from .api import (  # isort: skip
    convert,
    convert_progress,
    download,
    download_progress,
    load,
    process,
    process_progress,
    query,
)

# stupid pylance deleting my imports
# pylint: disable=pointless-statement
(
    convert,
    convert_progress,
    download,
    download_progress,
    load,
    process,
    process_progress,
    query,
)

__version__ = (1, 1, 1)
__version_str__ = ".".join(map(str, __version__))



from pathlib import Path
from typing import Iterator

from . import io, sources
from .comic import BaseComic
from .converter import ConvertFormats, get_converter
from .processor import ProcessOps, Processor


def query(url: str) -> BaseComic:
    """
    Attempt to query for a comic given a URL.
    :param `url`: An internet URL to search for
    :raises `ValueError` if the source is not found.
    """
    adapter = sources.get_class_for(url)(url)
    return BaseComic(adapter.metadata, adapter.chapters)


def load(path: Path | str) -> BaseComic:
    """
    Load a mandown-created comic from the file system.
    :param `path`: A folder where mandown has created a comic
    :throws FileNotFoundError if `md-metadata.json` cannot be found.
    """
    return io.read_comic(path)


def save_metadata(comic: BaseComic, path: Path | str) -> None:
    """
    Save the metadata from the comic to `<path>/md-metadata.json`.
    """
    io.save_comic(comic, path)


def init_parse_comic(path: Path | str, source_url: str | None = None) -> BaseComic:
    """
    Open a comic from a folder path, either via `md-metadata.json` or
    if that fails, parse the comic structure and create an `md-metadata.json`

    :param `path`: A folder containing `md-metadata.json` or a comic structure
    :param `source_url`: A source URL to fill metadata from
    :returns A comic with metadata and chapter data of that folder
    """
    try:
        comic = io.read_comic(path)
    except FileNotFoundError:
        comic = io.parse_comic(path, source_url)
        io.save_comic(comic, path)
    return comic


def convert_progress(
    comic: BaseComic,
    folder_path: Path | str,
    convert_to: ConvertFormats,
    dest_folder: Path | str | None = None,
) -> Iterator[str]:
    """
    From metadata in `comic`, convert the comic in `folder_path`
    to `convert_to` and put it in `dest_folder` (defaults to workdir).

    Returns an Iterator representing a progress bar up to the
    number of chapters in the comic.
    """
    if convert_to == ConvertFormats.NONE:
        return

    # default to working directory
    dest_folder = dest_folder or Path(".").resolve()

    # obviously pylint is wrong because this is 100% callable
    converter = get_converter(convert_to)(comic)  # pylint: disable=not-callable
    yield from converter.create_file_progress(folder_path, dest_folder)


def convert(
    comic: BaseComic,
    folder_path: Path | str,
    convert_to: ConvertFormats,
    dest_folder: Path | str | None = None,
) -> None:
    """
    From metadata in `comic`, convert the comic in `folder_path`
    to `convert_to` and put it in `dest_folder` (defaults to workdir).
    """
    for _ in convert_progress(comic, folder_path, convert_to, dest_folder):
        pass


def process_progress(comic_path: Path | str, ops: list[ProcessOps]) -> Iterator[str]:
    """
    Process the comic in `comic_path` with `ops` in the order provided.

    Returns an Iterator representing a progress bar up to the
    number of chapters in the comic.
    """
    data = io.discover_local_images(comic_path)
    for _, images in data.items():
        for i in images:
            Processor(i).process(ops)
        yield "Processing"


def process(comic_path: Path | str, ops: list[ProcessOps]) -> None:
    """
    Process the comic in `comic_path` with `ops` in the order provided.
    """
    for _ in process_progress(comic_path, ops):
        pass


def download_progress(
    comic: BaseComic | str,
    path: Path | str = ".",
    *,
    start: int | None = None,
    end: int | None = None,
    threads: int = 4,
    only_download_missing: bool = True,
) -> Iterator[str]:
    """
    Download comic or comic URL `comic` to `path` using `threads` threads.

    If `start` or `end` are specified, only download those
    chapters (one-indexed).

    If `only_download_missing` is `True`, images already in the
    destination path will not be downloaded again.

    Returns an Iterator representing a progress bar up to the
    number of chapters in the comic.
    """
    path = Path(path)

    # make var comic a BaseComic
    if isinstance(comic, str):
        comic = query(comic)

    # create dir
    try:
        full_path = path / comic.metadata.title
    except IOError:
        # invalid filename
        full_path = path / comic.metadata.title_slug
    full_path.mkdir(exist_ok=True)

    # save metadata json
    comic.set_chapter_range(start=start, end=end)
    io.save_comic(comic, full_path)

    # cover
    if comic.metadata.cover_art:
        for _ in io.download_images(
            [comic.metadata.cover_art], full_path, filestems=["cover"]
        ):
            pass

    # for each chapter
    for chap in comic.chapters[start:end]:
        image_urls = comic.get_chapter_image_urls(chap)
        chapter_path = full_path / chap.slug
        chapter_path.mkdir(exist_ok=True)

        # expect that they're named by numbers only
        skip_images: set[int] = set()
        if only_download_missing:
            for file in chapter_path.iterdir():
                if file.stem == file.stem.rjust(io.NUM_LEFT_PAD_DIGITS, "0"):
                    try:
                        skip_images.add(int(file.stem))
                    except ValueError:
                        # expected if not an image file
                        pass

        if not image_urls or len(skip_images) == len(image_urls):
            # move to next chapter if there's nothing to download for this one
            yield chap.title
            continue

        # name them 00001.png, 00002.png, etc
        # skipping ones that already exist
        processed_image_urls, filestems = zip(
            *(
                (link, str(i).rjust(io.NUM_LEFT_PAD_DIGITS, "0"))
                for i, link in enumerate(image_urls, start=1)
                if i not in skip_images
            )
        )

        chapter_path = full_path / chap.slug

        for _ in io.download_images(
            processed_image_urls,
            chapter_path,
            headers=comic.source.headers,
            filestems=filestems,
            threads=threads,
        ):
            pass
        yield chap.title


def download(
    comic: BaseComic | str,
    path: Path | str = ".",
    *,
    start: int | None = None,
    end: int | None = None,
    threads: int = 4,
    only_download_missing: bool = True,
) -> None:
    """
    Download comic or comic URL `comic` to `path` using `threads` threads.

    If `start` or `end` are specified, only download those
    chapters (one-indexed).

    If `only_download_missing` is `True`, images already in the
    destination path will not be downloaded again.
    """
    for _ in download_progress(
        comic,
        path,
        start=start,
        end=end,
        threads=threads,
        only_download_missing=only_download_missing,
    ):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mandown',
 'mandown.converter',
 'mandown.processor',
 'mandown.sources',
 'mandown.ui']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.10.0,<5.0.0',
 'feedparser>=6.0.8,<7.0.0',
 'filetype>=1.1.0,<2.0.0',
 'lxml>=4.7.1,<5.0.0',
 'natsort>=8.1.0,<9.0.0',
 'python-slugify>=6.1.2,<7.0.0',
 'requests>=2.27.0,<3.0.0',
 'typer>=0.6.0,<0.7.0']

extras_require = \
{'gui': ['PySide6>=6.4.0,<7.0.0'], 'postprocessing': ['Pillow>=9.0.1,<10.0.0']}

entry_points = \
{'console_scripts': ['mandown = mandown.cli:main',
                     'mandown-gui = mandown.ui.ui:main']}

setup_kwargs = {
    'name': 'mandown',
    'version': '1.1.1',
    'description': 'Comic/manga/webtoon downloader and converter to CBZ/EPUB/PDF',
    'long_description': '# mandown\n\n![Supported Python versions](https://img.shields.io/pypi/pyversions/mandown)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![Download from PyPI](https://img.shields.io/pypi/v/mandown)](https://pypi.org/project/mandown)\n[![Download from the AUR](https://img.shields.io/aur/version/mandown-git)](https://aur.archlinux.org/packages/mandown-git)\n[![Latest release](https://img.shields.io/github/v/release/potatoeggy/mandown?display_name=tag)](https://github.com/potatoeggy/mandown/releases/latest)\n[![License](https://img.shields.io/github/license/potatoeggy/mandown)](/LICENSE)\n\nComic downloader and converter to CBZ, EPUB, and/or PDF as a Python library and command line application.\n\n## Supported sites\n\nTo request a new site, please file a [new issue](https://github.com/potatoeggy/mandown/issues/new).\n\n- https://mangasee123.com\n- https://manganato.com\n- https://webtoons.com\n- https://mangadex.org\n- https://mangakakalot.com\n- https://readcomiconline.li\n\n## Installation\n\nInstall the package from PyPI:\n\n```\npip3 install mandown\n```\n\nInstall the optional large dependencies for some features of Mandown:\n```\n# image processing\npip3 install Pillow\n\n# graphical interface (GUI)\npip3 install PySide6\n```\n\nArch Linux users may also install the package from the [AUR](https://aur.archlinux.org/packages/mandown-git.git):\n\n```\ngit clone https://aur.archlinux.org/mandown-git.git\nmakepkg -si\n```\n\nOr, to build from source:\n\nMandown depends on [poetry](https://github.com/python-poetry/poetry) for building.\n\n```\ngit clone https://github.com/potatoeggy/mandown.git\npoetry install\npoetry build\npip3 install dist/mandown*.whl\n```\n\n## Usage\n\n```\nmandown get <URL>\n```\n\nTo convert the download contents to CBZ/EPUB, append the `--convert` option. To apply image processing to the downloaded images, append the `--process` option.\n\n```\nmandown get <URL> --convert epub --process rotate_double_pages\n```\n\nTo download only a certain range of chapters, append the `--start` and/or `--end` options.\n\n> **Note:** `--start` and `--end` are *inclusive*, i.e., using `--start 2 --end 3` will download chapters 2 and 3.\n\nTo convert an existing folder without downloading anything (like a stripped-down version of https://github.com/ciromattia/kcc), use the `convert` command.\n\n```\nmandown convert <FORMAT> <PATH_TO_FOLDER>\n```\n\nTo process an existing folder without downloading anything, use the `process` command.\n\n```\nmandown process <PROCESS_OPERATIONS> <PATH_TO_FOLDER>\n```\n\nWhere `PROCESS_OPERATIONS` is an option found from running `mandown process --help`.\n\nRun `mandown --help` for more info.\n\n## Library usage\n\nTo just download the images:\n```python\nimport mandown\n\nmandown.download("https://comic-site.com/the-best-comic")\n```\n\nTo download and convert to EPUB:\n```python\nimport mandown\n\ncomic = mandown.query("https://comic-site.com/the-best-comic")\nmandown.download(comic)\nmandown.convert(comic, comic.metadata.title, "epub")\n```\n\nMore advanced stuff:\n\n```python\nimport mandown\n\n# load a previously downloaded comic...\ncomic = mandown.load("path/to/comic/folder")\nprint(comic.metadata, comic.chapters)\n\n# ...OR fetch a comic from a source by URL\ncomic = mandown.query(url_to_comic)\nprint(comic.metadata, comic.chapters)\n\n# download the first 10 chapters of comic to ./comic using 4 threads\nmandown.download(comic, "./comic/", threads=4, end=10)\n\n# apply image post-processing to comic images in ./comic\nmandown.process("./comic/", options=["rotate_double_pages", "trim_borders"])\n\n# convert comic located in ./comic to epub, storing it in ./epubs\nmandown.convert(comic, "./comic/", "epub", "./epubs")\n```\n',
    'author': 'Daniel Chen',
    'author_email': 'danielchen04@hotmail.ca',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/potatoeggy/mandown',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)



from dataclasses import dataclass

from slugify import slugify


@dataclass(slots=True)
class BaseMetadata:
    title: str
    authors: list[str]
    url: str
    genres: list[str]
    description: str
    cover_art: str
    title_slug: str = ""

    def __post_init__(self) -> None:
        self.title_slug = slugify(self.title)

    def asdict(self) -> dict:
        return {
            "title": self.title,
            "authors": self.authors,
            "url": self.url,
            "genres": self.genres,
            "description": self.description,
            "cover_art": self.cover_art,
        }


@dataclass(slots=True)
class BaseChapter:
    title: str
    url: str
    slug: str = ""

    def __post_init__(self) -> None:
        if not self.slug:
            self.slug = slugify(self.title)

    def asdict(self) -> dict:
        return {
            "title": self.title,
            "url": self.url,
            "slug": self.slug,
        }



from . import cbz, epub, pdf
from .base_converter import BaseConverter, ConvertFormats


def get_converter(convert_to: ConvertFormats) -> type[BaseConverter]:
    match convert_to:
        case ConvertFormats.CBZ:
            return cbz.CbzConverter
        case ConvertFormats.EPUB:
            return epub.EpubConverter
        case ConvertFormats.PDF:
            return pdf.PdfConverter
        case _:
            raise NotImplementedError(
                f"{convert_to} conversion has not been implemented yet!"
            )
